from .tonal_system import TonalSystem
from .tonal_system_element import TonalSystemElement
from .scale import Scale
from .gcycle import GCycle
from .balzano_diagram import BalzanoDiagram
from .utils import transpose_to_octave


import sys
import os
import glob
import numbers

def check_or_create_folder(name):
    if os.path.isdir(name):
        return
    os.makedirs(name + "/")

def transpose_to_octave(pitch_or_pitches, octave):
    if isinstance(pitch_or_pitches, numbers.Number):
        return (octave)*12 + pitch_or_pitches
    elif isinstance(pitch_or_pitches, list):
        return [(octave)*12 + p for p in pitch_or_pitches]
    


from .tonal_system_element import TonalSystemElement
from .scale import Scale
from typing import Union
import copy
import numpy as np
import matplotlib.pyplot as plt

class GCycle:
    def __init__(self, generator: TonalSystemElement):
        assert generator.is_generator(), 'GCycle must be initialized with a element that is a generator of the given system'
        self.generator = generator
        self.elements = self.generate_cycle()
        #self.scale_sizes = sorted([self.generator.inverse().pitch_class, self.generator.inverse().symmetrical().pitch_class])

    def generate_cycle(self):
        elements = [TonalSystemElement(0, self.generator.module)]
        for _ in range(self.generator.module - 1):
            elements.append(elements[-1] + self.generator)
        return elements

    #TODO: Gerar escala diatônica e pentatônica, dependendo das características
    #TODO: Gerar no modo dórico generalizado
    def diatonic_scale(self, tonic: int):
        length = self.generator.inverse().pitch_class

        start_element = TonalSystemElement(-self.generator.pitch_class, self.generator.module)
        
        sc_elements = [start_element]
        for _ in range(length-1):
            sc_elements.append(sc_elements[-1]+self.generator)
        sc_elements.sort()
        sc_elements.append(sc_elements[0])

        struct = tuple((sc_elements[i]-sc_elements[i-1]).pitch_class for i in range(1, len(sc_elements)))

        return Scale(self.generator.module, struct, tonic=tonic, name="Diatonic Scale")

    def next(self, elem: Union[TonalSystemElement, int], steps: int):
        real_elem = TonalSystemElement(elem, self.generator.module) if isinstance(elem, int) else elem
        next_index = (self.elements.index(real_elem) + steps) % len(self.elements)
        return copy.deepcopy(self.elements[next_index])

    def show(self):
        r = 10
        angle = 2 * np.pi / self.generator.module
        x = [self.elements[0].pitch_class]
        y = [r]

        figure, axes = plt.subplots()
        cycle = plt.Circle((0, 0), r, fill=False)
        axes.add_artist(cycle)

        for i in range(1, self.generator.module + 1):
            x.append(r * np.sin(i * angle))
            y.append(r * np.cos(i * angle))

        #plt.plot(x, y, markersize=20)
        plt.plot(x, y, 'wo', markersize=20)
        for i in range(self.generator.module):
            plt.text(x[i], y[i],self.elements[i], ha='center', va='center')
        plt.axis('equal')
        plt.axis('off')
        plt.show()

    def __str__(self):
        resp_str = f'\n*************Cycle:***************\n'
        resp_str += f'Elements: {[str(e) for e in self.elements]}\n'
        #resp_str += f'Scale Sizes: {self.scale_sizes[0]} {self.scale_sizes[1]}'
        return resp_str


from .tonal_system_element import TonalSystemElement
from .utils import check_or_create_folder
import copy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches


class Scale:
    def __init__(self, system_size: int, interval_struct: tuple, tonic=0, name="Generic Scale"):
        self.system_size = system_size
        self.interval_struct = interval_struct
        self.tonic = tonic
        self.elements = self.build_elements(tonic)
        self.name = name
        self.interval_vector = self.vector()
    
    def build_elements(self, tonic: int):
        elements = []
        actual = tonic
        for i in self.interval_struct:
            elements.append(TonalSystemElement(actual, self.system_size))
            actual += i
        return elements

    def set_tonic(self, tonic: int):
        self.tonic = tonic
        self.elements = self.build_elements(tonic)

    # TODO central_note
    def next(self, elem: int, steps: int):
        if isinstance(elem, int):
            real_elem = TonalSystemElement(elem, self.system_size)
            octave = int(elem // self.system_size)
        else:
            real_elem = elem
            octave = 0

        next_index = (self.elements.index(real_elem) + steps) % len(self.elements)
        if steps > 0 and self.elements[next_index].pitch_class < real_elem.pitch_class:
            octave += 1
        elif steps < 0 and self.elements[next_index].pitch_class > real_elem.pitch_class:
            octave -= 1
        return self.elements[next_index].pitch_class + (octave * self.system_size)

    # TODO: usar algoritmo de Manacher para otimizar
    def find_symmetric_rotation(self):
        struct = list(self.interval_struct)
        for i in range(0, len(struct)):
            rotated = struct[i:] + struct[:i]
            if rotated == rotated[::-1]:
                return i
        return -1

    # TODO: garantir que file_name tenha .scl e que o kbm substitua.
    def export_scala_files(self, file_name: str, kbm_pattern=None):
        chroma_cents = 1200 / self.system_size
        check_or_create_folder('scala_files')
        f = open(f'scala_files/{file_name}', 'w')

        # Headers
        f.write(f'! {file_name}\n!\n {self.name}\n {len(self.elements)}\n!')

        sum_interval = 0
        for interval in self.interval_struct:
            sum_interval += interval
            format_value = "{:.5f}".format(sum_interval * chroma_cents)
            f.write(f'\n {format_value}')
        f.close()

        kbm_name = file_name[:-3] + 'kbm'
        kbm = open(f'scala_files/{kbm_name}', 'w')

        if kbm_pattern == None:
            kbm_pattern = [i for i in range(len(self.elements))]

        if len(kbm_pattern) <= 12:
            length = 12
            kbm.write(
                f'! {kbm_name}\n{length}\n{0}\n{127}\n{60 + self.elements[0].pitch_class}\n{69}\n440.00000\n{len(self.elements)}\n')
            kbm.write('! Mapping.')
            for e in (kbm_pattern + ['x' for _ in range(12 - len(kbm_pattern))]):
                kbm.write(f'\n{e}')

        else:
            length = len(kbm_pattern)
            kbm.write(
                f'! {kbm_name}\n{length}\n{0}\n{127}\n{60 + self.elements[0].pitch_class}\n{69}\n440.00000\n{len(self.elements)}\n')
            kbm.write('! Mapping.')
            for e in kbm_pattern:
                kbm.write(f'\n{e}')

        kbm.close()

    def show(self):
        r = 5
        angle = 2 * np.pi / self.system_size
        i = 0
        ascending_chromatic = [i for i in range(self.system_size)]
        x = [ascending_chromatic[0]]
        y = [r]

        for i in range(1, self.system_size + 1):
            x.append(r * np.sin(i * angle))
            y.append(r * np.cos(i * angle))

        # plt.plot(x, y, 'wo', markersize=20)

        figure, axes = plt.subplots()
        cycle = plt.Circle((0, 0), r, fill=False)
        axes.add_artist(cycle)

        x_line = []
        y_line = []
        #        for i in range(self.system_size):
        for i in self.elements:
            x_line.append(x[i.pitch_class])
            y_line.append(y[i.pitch_class])
            plt.text(x[i.pitch_class], y[i.pitch_class], i.pitch_class, ha='center', va='center')

        x_line.append(x[0])
        y_line.append(y[0])
        plt.plot(x_line, y_line, markersize=20)
        plt.plot(x_line, y_line, 'wo', markersize=20)
        plt.axis('equal')
        plt.axis('off')

        plt.show()

    def vector(self):
        v = [0 for _ in range(int(self.system_size / 2))]
        scale = self.elements
        for i, pivot in enumerate(scale):
            for el in scale[i + 1:]:
                dist = min((el - pivot).pitch_class % self.system_size, (pivot - el).pitch_class % self.system_size)
                v[dist - 1] += 1
        return v

    def get_midi_pitch_classes(self):
        return [e.midi for e in self.elements]

    def get_elements(self):
        return [e.pitch_class for e in self.elements]

    def __eq__(self, o):
        if not isinstance(o, Scale):
            return False
        return (self.interval_struct == o.interval_struct) and (self.system_size == o.system_size)

    def __len__(self):
        return len(self.elements)

    def __str__(self):
        output_str = self.name
        output_str += f'\nElements: {[e.pitch_class for e in self.elements]}'
        output_str += f'\nInterval Vector: {self.interval_vector}'
        output_str += f'\nInterval Struct: {self.interval_struct}\n'
        return output_str



from .tonal_system_element import TonalSystemElement
from .gcycle import GCycle
import matplotlib.pyplot as plt
import numpy as np

class BalzanoDiagram:
    def __init__(self, system_size: int, x: TonalSystemElement, y: TonalSystemElement):
        self.system_size = system_size
        self.thirds = (x, y)
        self.scale = GCycle(x+y).diatonic_scale(0)
        self.matrix = self.build_matrix(system_size, x, y)
        self.compact_matrix = self.build_compact_matrix(system_size, x, y)
        self.dims = (len(self.matrix), len(self.matrix[0]))

    def build_compact_matrix(self, system_size: int, x: TonalSystemElement, y: TonalSystemElement):
        matrix = []
        gen = x+y
        #TODO: classe subgrupo??
        dims = [int((gen.inverse().pitch_class+1)/2) +1, int((gen.inverse().pitch_class-1)/2)+1]


        rotation = self.scale.find_symmetric_rotation()
        initial = self.scale.elements[rotation] if rotation!=-1 else self.scale.elements[0]
        init_pos = (initial*gen.inverse()).pitch_class
        coords = (init_pos%x.subgroup(), init_pos%y.subgroup())
        for i in range(coords[0], coords[0]+dims[0]):
            row = []
            for j in range(coords[1], coords[1]+dims[1]):
                row.append(TonalSystemElement(i*x.pitch_class + j*y.pitch_class, system_size))
            matrix.append(row)
        return matrix

    def build_matrix(self, system_size: int, x: TonalSystemElement, y: TonalSystemElement):
        matrix = []
        gen = x+y
        #TODO: classe subgrupo??
        dims = [x.subgroup()+1, y.subgroup()+1]

        rotation = self.scale.find_symmetric_rotation()
        initial = self.scale.elements[rotation] if rotation!=-1 else self.scale.elements[0]
        init_pos = (initial*gen.inverse()).pitch_class
        coords = (init_pos%x.subgroup(), init_pos%y.subgroup())

        for i in range(coords[0], coords[0]+dims[0]):
            row = []
            for j in range(coords[1], coords[1]+dims[1]):
                row.append(TonalSystemElement(i*x.pitch_class + j*y.pitch_class, system_size))
            matrix.append(row)
        return matrix

    def contains_scale(self):
        pos = [0,0]
        matrix = self.matrix
        actual = matrix[pos[0]][pos[1]]
        region = []
        step = 0
        while len(region)!=len(self.scale.elements)+1:
            region.append(actual)
            pos[step]+=1
            step = (step+1)%2
            actual = matrix[pos[0]%(len(matrix)-1)][pos[1]%(len(matrix[0])-1)]

        if region[-1]!=region[0]: return False
        region = sorted(region[:-1])
        region.append(region[0])

        r_struct = list((region[i]-region[i-1]).pitch_class for i in range(1, len(region)))
        s_struct = list(self.scale.interval_struct)
        for i in range(len(s_struct)):
            if (r_struct[i:]+r_struct[:i]) == s_struct: return True

        return False
    
    def show(self, compact=True):
        fig, ax = plt.subplots()

        matrix = self.matrix if not compact else self.compact_matrix
        dims = [len(matrix), len(matrix[0])]

        if self.contains_scale():

            #Paddings
            pad_x = 0.02*(dims[1]+1)
            pad_y = 0.021*(dims[0]+1)
            p = (pad_x+pad_y)/2

            #i and j handle matrix; x and y handle the coordinates
            i, j = 0, 0
            x, y = 0.5, 0.5

            #First step is vertical TODO: checar qual é o primeiro passo com base no tipo da escala
            h = False

            #Draw matrix lines till it reaches the start element again
            while (matrix[i][j]!=matrix[0][0] or (i,j)==(0,0)):
                #If step is horizontal, increments column; else increments row
                dest = (i,(j+1)%dims[1]) if h else ((i+1)%dims[0], j)
                pos_dest = (0.5+dest[1], 0.5+dest[0])

                #Compute paddings and color
                px = pad_x if h else 0
                py = 0 if h else pad_y
                color = 'r' if h else 'b'

                #Draw line and diagonal if needed
                if (dest[0]<dims[0] and dest[1]<dims[1]) and (dest[0]>=i and dest[1]>=j):
                    plt.plot([x+px, pos_dest[0]-px], [y+py, pos_dest[1]-py], color=color)
                    if (i+1 < dims[0]) and (j+1 < dims[1]) and (matrix[dest[0]][dest[1]]!=matrix[0][0]):
                        plt.plot([x+p, x+1-p], [y+p, y+1-p], color='g')
                    h = not h


                if dest[0]<i:
                    plt.plot([x-1+pad_x, x-pad_x], [0.5, 0.5], color='r')
                    plt.plot([x-1+p, x-p], [0.5+p, 1.5-p], color='g')
                elif dest[1]<j:
                    plt.plot([0.5, 0.5], [y-pad_y, y-1+pad_y], color='b')
                    plt.plot([0.5+p, 1.5-p], [y-1+p, y-p], color='g')

                i, j = dest
                x, y = pos_dest


        for i in range(dims[0]):
            for j in range(dims[1]):
                el = matrix[i][j]
                ax.text(j+0.5, i+0.5, str(el), va='center', ha='center')

        ax.set_xlim(0, dims[1])
        ax.set_ylim(0, dims[0])
        ax.set_xticks(np.arange(dims[1]))
        ax.set_yticks(np.arange(dims[0]))

        plt.axis('off')

        ax.grid()
        plt.show()

    def __str__(self):
        return str([[str(x) for x in row] for row in self.matrix])



import math

class TonalSystemElement:
    def __init__(self, pitch: int, module: int):
        self.pitch_class = pitch % module
        self.module = module

        self.cents = self.pitch_class * (1200/self.module)
        self.midi = (self.cents/100)
    
    def is_generator(self):
        return math.gcd(self.pitch_class, self.module)==1

    def __add__(self, o):
        assert isinstance(o, TonalSystemElement), f"Cannot add TonalSystemElement to {type(o)}"
        assert self.module == o.module, "Cannot add elements of different modules"
        return TonalSystemElement(self.pitch_class + o.pitch_class, self.module)

    def __sub__(self, o):
        assert isinstance(o, TonalSystemElement), f"Cannot subtract TonalSystemElement to {type(o)}"
        assert self.module == o.module, "Cannot subtract elements of different modules"
        return TonalSystemElement(self.pitch_class - o.pitch_class, self.module)

    def __mul__(self, o):
        assert isinstance(o, TonalSystemElement), f"Cannot multiply TonalSystemElement to {type(o)}"
        assert self.module == o.module, "Cannot multiply elements of different modules"
        return TonalSystemElement(self.pitch_class * o.pitch_class, self.module)

    def __eq__(self, o):
        return isinstance(o, TonalSystemElement) and self.pitch_class == o.pitch_class and self.module == o.module

    def __gt__(self, o):
        return isinstance(o, TonalSystemElement) and (self.pitch_class > o.pitch_class) and self.module == o.module

    def __lt__(self, o):
        return isinstance(o, TonalSystemElement) and self.pitch_class < o.pitch_class and self.module == o.module

    def __le__(self, o):
        return isinstance(o, TonalSystemElement) and self.pitch_class <= o.pitch_class and self.module == o.module

    def __ge__(self, o):
        return isinstance(o, TonalSystemElement) and self.pitch_class >= o.pitch_class and self.module == o.module

    def inverse(self):
        if math.gcd(self.pitch_class, self.module) != 1:
            return None
        else:
            for i in range(self.module):
                if (i * self.pitch_class) % self.module == 1:
                    return TonalSystemElement(i, self.module)

    def symmetrical(self):
        return TonalSystemElement(-self.pitch_class, self.module)

    def subgroup(self):
        i = 1
        pitch_class = self.pitch_class
        while pitch_class!=0:
            i+=1
            pitch_class = (pitch_class+self.pitch_class)%self.module
        return i
    
    def __str__(self):
        return str(self.pitch_class)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    readme = fh.read()

setup(name='edopi',
    version='0.0.1',
    url='https://github.com/luansimoes/edopi',
    license='MIT License',
    author='Flavia Elias e Luan Simões',
    long_description=readme,
    long_description_content_type="text/markdown",
    author_email='luansimoes@id.uff.br',
    keywords='microtonal edo music group-theoretic',
    description='Python package to deal with Group-Theoretic microtonal music structures',
    packages=find_packages(include=['edopi', 'edopi.*']),
    install_requires=['numpy','matplotlib'],)



from .gcycle import GCycle
from .tonal_system_element import TonalSystemElement
from .scale import Scale
from .balzano_diagram import BalzanoDiagram
from typing import Union
import math

class TonalSystem:
    def __init__(self, n: int, g=1):
        assert n>1, "Tonal System must have more than one element"
        self.cardinality = n
        self.generator = TonalSystemElement(g, n)
        self.cycle = GCycle(self.generator)


    def scale(self, elements=[], struct=[], name='Generic Scale'):
        assert (len(elements)==0) ^ (len(struct)==0), "argument must be either elements or struct"
        if len(elements)!=0:
            elements.sort()
            if all(isinstance(e, TonalSystemElement) for e in elements):
                circle_elements = elements + [elements[0]]
            elif all(isinstance(e, int) for e in elements):
                circle_elements = [TonalSystemElement(e, self.cardinality) for e in elements] + [TonalSystemElement(elements[0], self.cardinality)]

            struct = tuple((circle_elements[i]-circle_elements[i-1]).pitch_class for i in range(1, len(circle_elements)))

        return Scale(self.cardinality, struct, name=name)

    def diatonic_scale(self):
        return self.cycle.diatonic_scale(0)

    def get_generators(self):
        n = self.cardinality
        return [x for x in range(n) if math.gcd(x, n)==1]

    def set_generator(self, g: Union[TonalSystemElement, int]):
        if isinstance(g, int):
            assert math.gcd(g, self.cardinality)==1, "Element must be a generator"
            self.generator = TonalSystemElement(g, self.cardinality)
        elif isinstance(g, TonalSystemElement):
            assert g.is_generator(), "Element must be a generator"
            self.generator = g
        self.cycle = GCycle(self.generator)

    def get_midi_pitch_classes(self):
        n = self.cardinality
        return [TonalSystemElement(i, n).midi for i in range(n)]

    def midi_pitch(self, pitch_class: int, oct=0):
        oct+=2
        if oct==2:
            oct += int(pitch_class//self.cardinality) - 2
        return TonalSystemElement(pitch_class, self.cardinality).midi + (12*oct)

    def balzano_diagram(self, minor: int, major: int):
        n = self.cardinality
        assert TonalSystemElement(minor+major, n)==self.generator, "thirds must sum up to generator"
        return BalzanoDiagram(n, TonalSystemElement(minor, n), TonalSystemElement(major, n))

    def show_gCycle(self):
        self.cycle.show()
    
    def show(self):
        GCycle(TonalSystemElement(1, self.cardinality)).show()

    def __str__(self):
        return f'{self.cardinality}-Fold Tonal System with generator {self.generator}'
